﻿#from numba import jit, njit
#from numba.experimental import jitclass

from tkinter import *  #Tk, Canvas, Button, Frame, BOTH, NORMAL, HIDDEN
from tkinter import ttk
from random import Random
from time import sleep
from math import ceil

import matplotlib.pyplot as PLT
from time import time

#=====================================
#  GLOBALs and Settings
#
rnd  = Random()
root = None     # primary window
CV   = None     # canvas object

# ширина и высота окна
win_width = 800
win_height = 400

# ширина каждой клетки, с учетом просвета
cell_size = 8

# размеры поля в клетках
field_height = win_height//cell_size
field_width  = win_width //cell_size
half_width = field_width//2
cell_count   = field_height*field_width

# количество агентов
Agent_Density = 0.25    # "density" = agents per cell
Agent_Count = int(cell_count*Agent_Density)

Infected_Init_Part = 0.01   # which part of agents will be infected at time=0
Infected_Init_Count = ceil(Agent_Count*Infected_Init_Part)

# In fact, these are threshold values for random value probes.
# This means that AVERAGE value of probability is half of that!
#
RecoveryChance = 0.05   # daily probability of recovery for Infected  (I=>R)
InfectChance   = 0.25   # daily probability of infection for Susceptible (S=>I)
UnImmuneChance = 0.01   # daily probability of loosing Immunity for Recovered (R=>S)

# storage of agents:
Agents = [None]*Agent_Count

EMPTY_CELL = -1
DISABLED_CELL = -2

# matrix of cell occupation (who's occupied the cell)
Field = [[EMPTY_CELL]*field_width for y in range(field_height)]

Infected_Count  = 0
Recovered_Count = 0
Infected_History = []
Recovered_History= []

Infected_LCount  = 0
Recovered_LCount = 0
Infected_Left  = []
Recovered_Left = []

Infected_RCount  = 0
Recovered_RCount = 0
Infected_Right = []
Recovered_Right= []

L_R_BOUND = 58

time_start = 0
steps_delta = 0

COLORS = ["#C0C0C0", "#404040", "#FFFFFF", "red"]

VISUALIZE = True

#=====================================
# @ jitclass()
class Agent():
    #---
    def mybox(self):
        return (self.X*cell_size+1,           self.Y*cell_size+1,
                self.X*cell_size+cell_size-1, self.Y*cell_size+cell_size-1)
    
    #---
    def __init__(self, index, x0, y0):
        global CV, Field
        self.X = x0
        self.Y = y0
        self.State = "s"
        self.time = 0
        self.Index = index
        x1,y1,x2,y2 = self.mybox()
        self.obj = CV.create_rectangle(x1,y1,x2,y2, fill=COLORS[0])
        Field[self.Y][self.X] = self.Index

    #---
    def MoveTo(self, newX, newY):
        global CV, Field
        if newX==self.X and newY==self.Y: return
        if Field[newY][newX] == EMPTY_CELL:
            Field[newY][newX], Field[self.Y][self.X] = Field[self.Y][self.X], Field[newY][newX]
            if VISUALIZE:
                x1,y1,x2,y2 = self.mybox()
                CV.coords(self.obj, x1,y1,x2,y2)
            self.X, self.Y = newX, newY

    #---
    def ChangeState(self, newstate):
        global CV, Infected_Count, Recovered_Count
        global Infected_LCount, Recovered_LCount, Infected_RCount, Recovered_RCount

        if newstate=="s":
            if VISUALIZE:
                CV.itemconfig(self.obj, fill=COLORS[0])

            # change counters
            if self.State=="i":
                Infected_Count -= 1
                #
                if self.X<=L_R_BOUND:  Infected_LCount -= 1
                else:                  Infected_RCount -= 1
                
            elif self.State=="r":
                Recovered_Count -= 1
                #
                if self.X<=L_R_BOUND:  Recovered_LCount -= 1
                else:                  Recovered_RCount -= 1
                
            self.State = newstate
            self.time=0     # <<===

        elif newstate=="i":
            if VISUALIZE:
                CV.itemconfig(self.obj, fill=COLORS[1])

            # change counters
            if self.State=="r":
                Recovered_Count -= 1
                #
                if self.X<=L_R_BOUND: Recovered_LCount -= 1
                else:                 Recovered_RCount -= 1
                
            self.State = newstate
            self.time=1/RecoveryChance     # <<===
            Infected_Count += 1
            #
            if self.X<=L_R_BOUND:  Infected_LCount += 1
            else:                  Infected_RCount += 1
                

        elif newstate=="r":
            if VISUALIZE:
                CV.itemconfig(self.obj, fill=COLORS[2])
            
            # change counters
            if self.State=="i":
                Infected_Count -= 1
                #
                if self.X<=L_R_BOUND:  Infected_LCount -= 1
                else:                  Infected_RCount -= 1
                
            self.State = newstate
            self.time=1/UnImmuneChance     # <<===
            Recovered_Count += 1
            #
            if self.X<=L_R_BOUND:  Recovered_LCount += 1
            else:                  Recovered_RCount += 1

        else:  # it's some unidentified state: agent is marked by RED color.
            if VISUALIZE:
                CV.itemconfig(self.obj, fill=COLORS[3])

            # change counters
            if self.State=="i":
                Infected_Count -= 1
                #
                if self.X<=L_R_BOUND:  Infected_LCount -= 1
                else:                  Infected_RCount -= 1
                
            elif self.State=="r":
                Recovered_Count -= 1
                #
                if self.X<=L_R_BOUND:  Recovered_LCount -= 1
                else:                  Recovered_RCount -= 1
                
            self.State = newstate
            self.time=0     # <<===

#=====================================
# move agents by 1 cell at random direction 
# @ njit
def Walk():
    global Agents, Field

    # iterate through agents
    for i in range(len(Agents)):
        x = Agents[i].X
        y = Agents[i].Y

        # поиск свободных клеток по соседству
        free_neighbours = []
        if   x>0             and Field[y][x-1]==EMPTY_CELL: free_neighbours.append( (-1,0) )
        if x+1<len(Field[y]) and Field[y][x+1]==EMPTY_CELL: free_neighbours.append( (+1,0) )
        if   y>0             and Field[y-1][x]==EMPTY_CELL: free_neighbours.append( (0,-1) )
        if y+1<len(Field)    and Field[y+1][x]==EMPTY_CELL: free_neighbours.append( (0,+1) )

        if len(free_neighbours)>0:
            dx,dy = rnd.choice(free_neighbours) # случайный выбор направления
            Agents[i].MoveTo(x+dx,y+dy)

#=====================================
# agents can infect others and can be recovered
#
# @ njit
def Infect_And_Recover():
    global Agents, Field

    # iterate through agents
    for i in range(len(Agents)):
        if Agents[i].State=="i":  # the agent is infected
            Agents[i].time -= 1
            # the agent can be recovered
            #if rnd.random()<=RecoveryChance:
            if Agents[i].time<=0: # and rnd.random()<=RecoveryChance:
                Agents[i].ChangeState("r")
            else:
                # the agent can infect neighbours
                # список соседей
                x = Agents[i].X
                y = Agents[i].Y
                neighbours = []
                if x>0 and Field[y][x-1]>=0:                #!=EMPTY_CELL:
                    j = Field[y][x-1]
                    if Agents[j].State=="s":
                        neighbours.append(j)
                if x+1<len(Field[0]) and Field[y][x+1]>=0:  #!=EMPTY_CELL:
                    j = Field[y][x+1]
                    if Agents[j].State=="s":
                        neighbours.append(j)
                if y>0 and Field[y-1][x]>=0:                #!=EMPTY_CELL:
                    j = Field[y-1][x]
                    if Agents[j].State=="s":
                        neighbours.append(j)
                if y+1<len(Field) and Field[y+1][x]>=0:     #!=EMPTY_CELL:
                    j = Field[y+1][x]
                    if Agents[j].State=="s":
                        neighbours.append(j)
                        
                # try to infect neighbours
                for j in neighbours:
                    if rnd.random()<=InfectChance:
                        Agents[j].ChangeState("i")
        elif Agents[i].State=="r":  # the agent can loose immunity
            Agents[i].time -= 1
            #if rnd.random()<=UnImmuneChance:
            if Agents[i].time<=0: #and rnd.random()<=UnImmuneChance:
                Agents[i].ChangeState("s")

#                    

#=====================================
def Step():
    global Infected_History, Recovered_History, Infected_Count, Recovered_Count
    global Infected_Left, Recovered_Left, Infected_LCount, Recovered_LCount
    global Infected_Right, Recovered_Right, Infected_RCount, Recovered_RCount
    global steps_delta

    Infect_And_Recover()
    Walk()
    steps_delta += 1

    Infected_History.append( Infected_Count /Agent_Count*100)
    Recovered_History.append(Recovered_Count/Agent_Count*100)

    Infected_Left.append(  Infected_LCount /Agent_Count*100)
    Recovered_Left.append( Recovered_LCount/Agent_Count*100)
    Infected_Right.append( Infected_RCount /Agent_Count*100)
    Recovered_Right.append(Recovered_RCount/Agent_Count*100)

#=====================================
def Init():
    global Agents, Field
    global Agent_Density, Infected_Init_Part, InfectChance, RecoveryChance
    global UnImmuneChance, Agent_Count, Infected_Init_Count, cell_count
    global Infected_Count, Recovered_Count, Infected_History, Recovered_History
    global Infected_Left, Recovered_Left, Infected_LCount, Recovered_LCount
    global Infected_Right, Recovered_Right, Infected_RCount, Recovered_RCount
    global steps_delta, time_start

    Agent_Density       = float(sbDensity.get())/100
    Infected_Init_Part  = float(sbInfect0.get())/100
    InfectChance        = float(sbInfectProb.get())/100
    RecoveryChance      = float(sbRecoverProb.get())/100
    UnImmuneChance      = float(sbUnImmuneProb.get())/100

    #-- init map --
    
    # entire map is EMPTY
    #
    for y in range(field_height):
        for x in range(field_width):
            Field[y][x] = EMPTY_CELL

    # setup WALLS
    #

    #--
    # load map from file
    mapfile = open("map.txt").readlines()
    for y in range(min(len(mapfile), field_height)):
        for x in range(min(len(mapfile[y]), field_width)):
            Field[y][x] = DISABLED_CELL if mapfile[y][x]=="#" else EMPTY_CELL

    # update cell_count
    #
    cell_count = sum( Field[y].count(EMPTY_CELL) for y in range(field_height) )

    Agent_Count = int(cell_count*Agent_Density)
    Agents = [None]*Agent_Count

    Infected_Init_Count = ceil(Agent_Count*Infected_Init_Part)

    CV.delete("all")

    # create agents
    for i in range(len(Agents)):
        while True:
            x = rnd.randint(0, field_width-1)
            y = rnd.randint(0, field_height-1)
            if Field[y][x]==EMPTY_CELL:
                Agents[i] = Agent(i,x,y)
                break

    # choose infected agents
    for i in range(Infected_Init_Count):
        while True:
            j = rnd.randint(0, len(Agents)-1) 
            if Agents[j].State=="s":
                Agents[j].ChangeState("i")
                break

    # reset History
    Infected_Count  = 0
    Recovered_Count = 0
    Infected_History.clear()
    Recovered_History.clear()

    Infected_LCount  = 0
    Recovered_LCount = 0
    Infected_Left.clear()
    Recovered_Left.clear()
    Infected_RCount  = 0
    Recovered_RCount = 0
    Infected_Right.clear()
    Recovered_Right.clear()
    
    PLT.cla()   # clear plots and reset axes

    steps_delta = 0
    time_start = time()

#=====================================
FLAG_RUN = False

def Run(*args):
    global FLAG_RUN
    
    if not FLAG_RUN: return
    Step()
    root.after(1, Run)    # таймер на Питоне

#=====================================
def PrintTiming():
    global steps_delta, time_start

    time_end = time()
    elapsed = time_end-time_start
    print("steps={}, elapsed={:1.1f} [sec], rate={:1.2f} [step/sec]".format(steps_delta, elapsed, steps_delta/elapsed))
    return
    
#=====================================
def ToggleRun(*args):
    global FLAG_RUN, bnRun
    global steps_delta, time_start
    
    FLAG_RUN = not FLAG_RUN
    if FLAG_RUN:
        bnRun.configure(text="Stop")
        Run(args)
        steps_delta = 0
        time_start = time()
    else:
        bnRun.configure(text="Run")
        PrintTiming()

#=====================================
def SaveInfo(fname="history.txt"):

    global Agents, Field
    global Agent_Density, Infected_Init_Part, InfectChance, RecoveryChance
    global UnImmuneChance, Agent_Count, Infected_Init_Count, cell_count
    global Infected_Count, Recovered_Count, Infected_History, Recovered_History
    global Infected_Left, Recovered_Left, Infected_LCount, Recovered_LCount
    global Infected_Right, Recovered_Right, Infected_RCount, Recovered_RCount
    global steps_delta, time_start

    fmt="{}; {}; {}; {}; {}; {};\n"
    F = open(fname, "wt", encoding="utf-8")
    F.write(fmt.format("Infected","Recovered", "L_inf", "L_recv", "R_inf", "R_recv"))
    for k in range(min(map(len, [Infected_History, Recovered_History, Infected_Left, Recovered_Left, Infected_Right, Recovered_Right]))):
        I,R, IL,RL, IR,RR = Infected_History[k],Recovered_History[k], Infected_Left[k],Recovered_Left[k], Infected_Right[k], Recovered_Right[k]
        F.write(fmt.format(I,R, IL,RL, IR,RR))
    F.close()
    return

#=====================================
def ShowInfo(*args):
    global Counts_History
    global steps_delta, time_start

    PrintTiming()

    PLT.clf()
#    PLT.plot(Infected_History, color="red",  label="Infected", linewidth=1)
#    PLT.plot(Recovered_History,color="green",label="Recovered",linewidth=1)

    PLT.plot(Infected_Left,  color="red",  label="Infected Left",  linestyle="dashdot", linewidth=1)
    PLT.plot(Recovered_Left, color="green",label="Recovered Left", linestyle="dashdot", linewidth=1)
    PLT.plot(Infected_Right, color="red",  label="Infected Right", linestyle="dashed",  linewidth=1)
    PLT.plot(Recovered_Right,color="green",label="Recovered Right",linestyle="dashed",  linewidth=1)

    PLT.legend()
    PLT.subplots_adjust(left=0.06,  right=0.99,
                        bottom=0.05,top=0.97,
                        wspace=0.1, hspace=0.1)
    PLT.draw()

    SaveInfo("history."+str(len(Infected_History))+".txt")
    
    try:        # suppress error messages when closing windows
        PLT.pause(1)
    except:
        pass
    steps_delta = 0
    time_start = time()

#=====================================
# @ njit
def OriginCoords(ix, iy):
    global cell_size
    return cell_size*ix, cell_size*iy


#=====================================
def ToggleVisual(*args):
    global VISUALIZE, sbVISU, Agents, Agent_Count
    #VISUALIZE = not VISUALIZE
    #= 1-VISUALIZE
    VISUALIZE = bool(sbVISU.get())

    # update entire view
    #
    if VISUALIZE:
        for ag in Agents:
            ag.ChangeState(ag.State)
            ag.MoveTo(ag.X, ag.Y)

#=====================================
#   MAIN()
#=====================================

# Создаем окно
root = Tk()

# Методом geometry() задаем размеры окна
config_string = "{0}x{1}".format(win_width+2, win_height + 148)
root.geometry(config_string)

# Создается канва (объект типа Canvas), на ней происходит рисование.
# Канва "присоединяется" к окну.
#
CV = Canvas(root, height=win_height, width=win_width)
CV.configure(bg="white")

# Выравнивание канвы в окне
#
CV.grid(row=0,column=0,rowspan=1,columnspan=4)

# создаём фрейм для элементов "управления" моделью
frMenu = Frame(root, relief=SUNKEN)
frMenu.grid(row=1, column=0, columnspan=3, pady=3)      # размещение элементов внутри фрейма = "сетка".

bnInit = Button(frMenu, text='Init', command = Init)    # создаём кнопку "Init"
bnInit.grid(row=0, column=0, padx=3)                    # и размещаем её на фрейме
bnStep = Button(frMenu, text='Step', command = Step) 
bnStep.grid(row=0, column=1, padx=3)
bnRun  = Button(frMenu, text='Run',  command = ToggleRun)
bnRun .grid(row=0, column=2, padx=3)
bnInfo = Button(frMenu, text='Info', command = ShowInfo)
bnInfo.grid(row=0, column=3, padx=3)

sbVISU = IntVar()                                       # создаём "флажок" (combobox)
sbVISU.set(int(VISUALIZE))                              # который будет отключать визуализацию движения агентов
sbVisu = Checkbutton(frMenu, text="(v)", variable=sbVISU, command = ToggleVisual)
sbVisu.grid(row=0, column=5, columnspan=1, padx=3)

# фрейм параметров
frParams = Frame(root)
frParams.grid(row=2, column=0, columnspan=4, pady=3)

param_Agent_Density = DoubleVar()
param_Infected_Init = DoubleVar()
param_InfectChance  = DoubleVar()
param_RecoverChance = DoubleVar()
param_UnImmuneChance= DoubleVar()

label1 = Label(frParams, text="Population density [%] ")
label1.grid(row=0, column=0, padx=3, sticky=W)
sbDensity = Spinbox(frParams, from_=0, to_=100, increment=0.01, width=8,
                    textvariable=param_Agent_Density)
sbDensity.grid(row=0, column=1, padx=3) #, sticky=E)

label2 = Label(frParams, text="Initially Infected [%] ")
label2.grid(row=1, column=0, padx=3)
sbInfect0 = Spinbox(frParams, from_=0, to_=100, increment=0.01, width=8,
                    textvariable=param_Infected_Init)
sbInfect0.grid(row=1, column=1, padx=3) #, sticky=E)

label3 = Label(frParams, text="Infect daily chance [%] ")
label3.grid(row=2, column=0, padx=3)
sbInfectProb = Spinbox(frParams, from_=0, to_=100, increment=0.01, width=8,
                    textvariable=param_InfectChance)
sbInfectProb.grid(row=2, column=1, padx=3) #, sticky=E)

label4 = Label(frParams, text="Recover daily chance [%] ")
label4.grid(row=3, column=0, padx=3)
sbRecoverProb = Spinbox(frParams, from_=0, to_=100, increment=0.01, width=8,
                    textvariable=param_RecoverChance)
sbRecoverProb.grid(row=3, column=1, padx=3) #, sticky=E)

label5 = Label(frParams, text="Un-immune daily chance [%] ")
label5.grid(row=4, column=0, padx=3)
sbUnImmuneProb = Spinbox(frParams, from_=0, to_=100, increment=0.01, width=8,
                    textvariable=param_UnImmuneChance)
sbUnImmuneProb.grid(row=4, column=1, padx=3) #, sticky=E)

param_Agent_Density .set( Agent_Density *100)
param_Infected_Init .set( Infected_Init_Part*100)
param_InfectChance  .set( InfectChance  *100)
param_RecoverChance .set( RecoveryChance*100)
param_UnImmuneChance.set( UnImmuneChance*100)

#--- создаем агентов ---
Init()

# стандартный цикл обработки cобытий,
# общая работа оконного приложения
root.mainloop()
#=====================================
